{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE LambdaCase #-}
{-# LANGUAGE OverloadedStrings #-}

module Lisa.Types where

import Control.Concurrent (MVar, newMVar, putMVar, takeMVar)
import Control.Monad.IO.Class (liftIO)
import Data.Text (Text)
import Data.Time.Clock (UTCTime, addUTCTime, getCurrentTime)
import Network.HTTP.Req (https)
import Text.Printf (printf)
import Web.Spock (SpockAction, getState)

import Lisa.Squeak (Lecture, SqueakCtx, SqueakError)

import qualified Lisa.Squeak as Squeak

-- | Session data per-client.
data Session = Session
    { sessionAuthToken :: Maybe Text
    } deriving (Show)

emtpySession :: Session
emtpySession = Session Nothing

-- | Global application state.
data State = State
    { stateLectures :: MVar (UTCTime, [Lecture]) -- ^ Cached lectures with cache expiration time.
    }

mkState :: IO State
mkState = do
    now <- getCurrentTime
    State <$> newMVar (now, [])

-- | Get lectures from cache if they've been retrieved before or from Squeak.
-- Uses 'MVar's for synchronisation: As long as each thread 'takeMVar's before it 'putMVar's, we're in the clear.
-- TODO: Evaluate lectures to NF before 'putMVar'ing them.
getAllLecturesCached :: SpockAction conn sess State (Either SqueakError [Lecture])
getAllLecturesCached = do
    lecturesMVar <- stateLectures <$> getState
    (expirationTime, lectures) <- liftIO $ takeMVar lecturesMVar
    now <- liftIO getCurrentTime
    (newContents, result) <- if now >= expirationTime
    then -- Cache expired => fetch lectures again.
        Squeak.getAllLectures >>= \case
            -- On error: Clear cache.
            Left err ->
                pure ((now, []), Left err)
            -- On success: Cache fetched lectures for 60 seconds.
            Right lectures' -> do
                let expirationTime' = addUTCTime 60 now
                liftIO $ printf "Fetched %d lectures, valid until %s\n" (length lectures') (show expirationTime')
                pure ((expirationTime', lectures'), Right lectures')
    else -- Cache still valid => return it.
        pure ((expirationTime, lectures), Right lectures)
    liftIO $ putMVar lecturesMVar newContents
    pure result

instance SqueakCtx (SpockAction conn sess State) where
    getLocationInfo = pure $ Right (https "api.squeak-test.fsmi.uni-karlsruhe.de", mempty)
    getAuthToken = pure Nothing
